#!/usr/bin/perl -w

package mysqlstat;

use strict;
use vars qw(@ISA @EXPORT @ARGV $VERSION);

use AppConfig qw(:expand :argcount);
use Fcntl qw(:DEFAULT :flock);

use DBI;
use Storable qw(retrieve_fd nstore_fd);
use RRDs;

$VERSION= '$Id: $';
@ISA = qw(Exporter);

@EXPORT = qw(
  &getConfig
  &printConfig
  &getRetentionData
  &saveRetentionData
  &getServerStatus
  &printServerStatus
  &createServerRRD
  &updateServerRRD
  &calculateDiff
  &processServer
	&formatUptime
	&formatTrafic
	&formatQuery
);

#	&processServers
#	&checkConfig
#	&checkServerDefinitions
#	&findConfigFile
#	&printCurrentConfig
#	&createServerRRD
#	&updateServerRRD
#	&getServerStatus
#	&loadData
#	&saveData
#	&unpackStatus
#	&packStatus


###############################################################################
###############################################################################
my %servers;
my $current_server;

sub getConfig {

	my ($verbose,$work_dir,$config_file);

	my $config = AppConfig->new({
			CREATE   => 1,
	#		CASE   	 => 1,

			GLOBAL   => {
				DEFAULT  => '',
				ARGCOUNT => ARGCOUNT_ONE,
			}

	});

	$config->define(
  	"work_dir|w=s"    => { ARGCOUNT => ARGCOUNT_ONE , DEFAULT => '' },
	  "config_file|c=s" => { ARGCOUNT => ARGCOUNT_ONE , DEFAULT => '' },
  	"test|t"          => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 0 },
	  "help|h"					=> { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 0, ACTION => \&printUsage },
  	"version|V"				=> { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 0, ACTION => \&printVersion },
		"verbose|d" 			=> { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 0, ACTION => sub { $verbose++; } },
		"server_disabled"	=> { ACTION 	=> \&parseServerConfig ,ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1},
		"server_uid"			=> { ACTION 	=> \&parseServerConfig },
		"server_name"			=> { ACTION 	=> \&parseServerConfig },
		"server_port"			=> { ACTION 	=> \&parseServerConfig },
		"server_host" 		=> { ACTION 	=> \&parseServerConfig },
		"server_user" 		=> { ACTION 	=> \&parseServerConfig },
		"server_password" => { ACTION 	=> \&parseServerConfig },
	);

	$config->getopt(\@ARGV);

	$work_dir = $config->work_dir;
	$config_file = $config->config_file;

	#-     
	if (!$config_file){
  	$config_file = findConfigFile();

	  printError("Can not find config file\n") unless($config_file);
	}

	$config->file($config_file) or printError("Can not open config file [$config_file] ($!)\n");

	$work_dir = $config->work_dir unless ($work_dir);
	$verbose = $config->verbose unless ($verbose);

	my %config = (
		config_file => $config_file,
		work_dir    => $work_dir,
		verbose     => $verbose,
		test        => $config->test,
		servers     => { %servers },
	);

	checkConfig(\%config);

	return \%config;
}


###############################################################################
###############################################################################
sub findConfigFile {
  my $config_file = '';

  my @dirs = qw(/etc/mysqlstat/mysqlstat.conf /etc/mysqlstat.conf);

  unshift (@dirs,$ENV{HOME}."/.mysqlstat.conf") if ($ENV{HOME});

  foreach(@dirs){
  	if ( -f $_ ){
  		$config_file = $_;
  		last;
  	}
  }

  return $config_file;
}

###############################################################################
###############################################################################
sub checkConfig {
  my $config = shift;

	printError("Please specified working dir\n") if (!$config->{work_dir});
	printError("Wrong working directory $config->{work_dir} ($!)\n") unless ( -d $config->{work_dir} );

	my ($perms,$oct_perms);
	$perms = (stat($config->{config_file}))[2] & 07777;
  $oct_perms = sprintf("%lo", $perms);

	printError("Your config file ($config->{config_file}) has wrong permissions ($oct_perms) ! Must be 600 !!!\n") if ($oct_perms ne "600");

	checkServerDefinitions($config);
}

###############################################################################
###############################################################################
sub checkServerDefinitions{
  my $config = shift;

	foreach(keys %{ $config->{servers} }){
		my $s = $config->{servers}->{$_};

		if (!exists($s->{host}) || !exists($s->{name}) || !exists($s->{user})) {
			printError("Not completed definition for [$_]\n");
		}
	}
}

###############################################################################
###############################################################################
sub parseServerConfig {
	my ($config,$varname,$value) = @_;

	print "parseServers() $varname => $value\n" if ($config->verbose>3);

	return 1 unless ($varname =~ /^server_/);

	if ($varname eq 'server_uid'){
		if (exists($servers{$value})){
			die("Server UID $value already used\n");
		}

		$current_server = $value;

		$servers{$current_server} = {
			uid => $current_server
		};

	}elsif(!$current_server){
		die("Unknown Server UID\n");
	}elsif ($varname =~ /^server_(.+)$/){
	  $servers{$current_server}->{$1} = $value;
	}

	return 1;
}

###############################################################################
###############################################################################
sub printConfig{
  my $config = shift;

	print "\n";
	print "Current config\n";
	print "="x80,"\n";
	print "Work_dir:    $config->{work_dir}\n";
	print "Config_file: $config->{config_file}\n";
	print "Verbose:     $config->{verbose}\n";
	print "\n";

	print "Server list\n";
	print "="x80,"\n";
	foreach(sort keys %{ $config->{servers} }){
 	  print "UID: $_\n";

		my $s = $config->{servers}->{$_};

		foreach(sort keys %{ $s }){
		  next if (/uid/);
			print "\t$_ => ",(/^password$/)?'*'x8:$s->{$_},"\n";
		}

 	  print "\n";
	}
}

sub printError{
	print @_;
	exit;
}

my $RetentionDataVer = 3;

sub getRetentionData {
	my $config = shift;
	my $cgi_mode = shift;

	my $file_name;
	my $rdata;

	if ($cgi_mode){
		$file_name = $config.'/mysqlstat.dat';
	}else{
		$file_name = $config->{work_dir}.'/mysqlstat.dat';
	}

	my $new = 1 unless ( -f $file_name );

  if ($new){
    return undef if ($cgi_mode);

    $rdata = {
      Version  				=> $RetentionDataVer,
	  	Started 				=> time,
  		LastUpdated			=> time,
	  	Total_Questions => 0,
  		Total_In				=> 0,
	  	Total_Out				=> 0,
	  	servers					=> {},
    };
  }else{
    open(DF, "< $file_name") or printError("Can't open retention file ($file_name): $!");
 	  flock(DF, LOCK_SH) or printError("Can't lock retention file ($file_name): $!");
	  $rdata = retrieve_fd(\*DF);
  	close(DF);

	  if (!defined($rdata->{Version}) || $rdata->{Version} != $RetentionDataVer){
  		printError("Wrong version of Retention data. Sorry can't continue :(\n");
	  }

	  return $rdata if ($cgi_mode);

		foreach(keys %{ $rdata->{servers} }){
			unless (exists($config->{servers}->{$_})){
				my $rrd_file = $config->{work_dir}.'/mysqlstat-'.$_.'.rrd';
				unlink($rrd_file) if ( -f $rrd_file);
				delete($rdata->{servers}->{$_});
				next;
			}
		}
	}

	foreach(keys %{ $config->{servers} }){

	  # New server
		unless (exists($rdata->{servers}->{$_})){
			$rdata->{servers}->{$_} = {};
			$rdata->{servers}->{$_}->{status} = undef;

			$rdata->{servers}->{$_}->{StartTime} = time;
			$rdata->{servers}->{$_}->{Restarts} = 0;
		}

		# Update from config
		$rdata->{servers}->{$_}->{name} = $config->{servers}->{$_}->{name};
		$rdata->{servers}->{$_}->{host} = $config->{servers}->{$_}->{host};
		$rdata->{servers}->{$_}->{disabled} = (exists($config->{servers}->{$_}->{disabled}) && $config->{servers}->{$_}->{disabled})+0;
  }

  return $rdata;
}

sub saveRetentionData {
	my $config = shift;
	my $rdata = shift;

  $rdata->{LastUpdated}	= time;

	my $file_name = $config->{work_dir}.'/mysqlstat.dat';

  sysopen(DF, "$file_name", O_RDWR|O_CREAT, 0644) or  printError("Can't open retention file ($file_name): $!");

  flock(DF, LOCK_EX) or printError("Can't lock retention file ($file_name): $!");

  nstore_fd($rdata, \*DF) or printError("Can't save retention data: $!");

  truncate(DF, tell(DF));

  close(DF);
}


###############################################################################
###############################################################################

my @RRD_DS = qw(
	Questions
	Slow_queries
	Connections
	Bytes_sent
	Bytes_received
	Max_used_connections
	Open_tables
	Open_files
	Com_select
	Com_insert
	Com_delete
	Com_update
	Com_change_db
	Created_tmp_disk_tables
	Created_tmp_tables
	Key_read_requests
	Key_reads
	Key_write_requests
	Key_writes
	Qcache_hits
	Qcache_inserts
	Qcache_not_cached
);

my %RRD_DS_TYPE = qw(
 	Questions						 		COUNTER
	Slow_queries 				 		COUNTER
	Connections 				 		COUNTER
	Bytes_sent 					 		COUNTER
	Bytes_received 			 		COUNTER
	Max_used_connections 		GAUGE
	Open_tables          		GAUGE
	Open_files           		GAUGE
	Com_select					 		COUNTER
	Com_insert           		COUNTER
	Com_delete           		COUNTER
	Com_update           		COUNTER
	Com_change_db        		COUNTER
	Created_tmp_disk_tables COUNTER
	Created_tmp_tables   		COUNTER
	Key_read_requests    		COUNTER
	Key_reads								COUNTER
	Key_write_requests			COUNTER
	Key_writes							COUNTER
	Qcache_hits							COUNTER
	Qcache_inserts					COUNTER
	Qcache_not_cached				COUNTER
);

my @RRD_RRA = (
	"RRA:AVERAGE:0.5:1:600",
	"RRA:AVERAGE:0.5:6:700",
	"RRA:AVERAGE:0.5:24:775",
	"RRA:AVERAGE:0.5:288:797",
	"RRA:MAX:0.5:1:600",
	"RRA:MAX:0.5:6:700",
	"RRA:MAX:0.5:24:775",
	"RRA:MAX:0.5:288:797",
	"RRA:LAST:0.5:1:600",
	"RRA:LAST:0.5:6:700",
	"RRA:LAST:0.5:24:775",
	"RRA:LAST:0.5:288:797",
);

sub createServerRRD{
  my $config = shift;
  my $server = shift;

  my $rrd_file = $config->{work_dir}.'/mysqlstat-'.$server->{uid}.'.rrd';

  my @params = ();

  foreach(@RRD_DS){
  	my $ds_name = substr($_,0,19);
  	my $ds_type = $RRD_DS_TYPE{$_};

  	push(@params,"DS:$ds_name:$ds_type:600:0:U");
  }

  foreach(@RRD_RRA){
  	push(@params,$_);
  }

  print "RRDs::create($rrd_file,".join(' ',@params)."\n" if ($config->{verbose}>2);

  RRDs::create($rrd_file,@params);

  my $ERR = RRDs::error;

  printError("Can't create RRD database ($ERR)\n") if ($ERR);
}

###############################################################################
###############################################################################
sub  updateServerRRD{
  my $config = shift;
  my $server = shift;
	my $status = shift;

  my $rrd_file = $config->{work_dir}.'/mysqlstat-'.$server->{uid}.'.rrd';

  my @ds = ();
	my @value = ('N');

  foreach(@RRD_DS){
 		my $ds_name = substr($_,0,19);

 		push(@ds,$ds_name);
 		push(@value, (defined($status) && exists($status->{$_})) ? $status->{$_} : 'U');
 	}

  print "RRDs::update($rrd_file,'-t',", join(':', @ds), ",", join(':', @value), ");\n" if ($config->{verbose}>2);

	RRDs::update($rrd_file,'-t',join(':',@ds),join(':',@value));

  my $ERR = RRDs::error;

  printError("Can't update RRD database ($ERR)\n") if ($ERR);
}

###############################################################################
###############################################################################
sub getServerStatus{
  my $server = shift;

  $server->{port} = 3306 unless(exists($server->{port}));

  my $dbh = DBI->connect("dbi:mysql::$server->{host}:$server->{port}",$server->{user},$server->{password}, { PrintError => 0, RaiseError => 0 });

  unless($dbh){
		print "Can not connect to $server->{user}\@$server->{host} ($DBI::errstr)\n";
		return undef;
  }

  my $q = $dbh->prepare('SHOW STATUS');

  $q->execute() or return undef;

  my %result;

  for(my $j=0;$j<$q->rows;$j++){
    my ($var_name,$value) = $q->fetchrow_array;

    $result{$var_name} = $value;
  }

  $q->finish();

  $q = $dbh->prepare("SHOW VARIABLES LIKE 'version'");
  $q->execute() or return undef;

  if ($q->rows){
		my ($var_name,$value) = $q->fetchrow_array;
		$result{Version} = $value;
	}else{
		$result{Version} = 'UNKNOWN';
	}

  $q->finish();

  $dbh->disconnect();

  $dbh = undef;
  $q = undef;

  return \%result;
}

sub processServer{
  my $config = shift;
  my $rdata = shift;
	my $server = shift;

	my $rserver = $rdata->{servers}->{$server->{uid}};

	print "Server: $server->{name} ($server->{uid})\n" if ($config->{verbose});

	my $rrd_file = $config->{work_dir}.'/mysqlstat-'.$server->{uid}.'.rrd';

  unless ( -f $rrd_file){
    print "RRD file [$rrd_file] was not found, Trying to create it ....\n" if ($config->{verbose}>2);
    createServerRRD($config,$server);
  }

	my $server_status = getServerStatus($server);

	printServerStatus($server_status)  if ($config->{verbose});

	unless(defined($server_status)){
		print "Can not get current server status\n" if ($config->{verbose});

		updateServerRRD($config,$server,$server_status);

		$rserver->{status} = 0;
		$rserver->{LastUpdate} = time;

		return;
	}

	# New server ?
	unless (defined($rdata->{servers}->{$_}->{status})){
    print "New server\n\n" if ($config->{verbose});

		updateServerRRD($config,$server,$server_status);

		$rserver->{status} = $server_status;
		$rserver->{LastUpdate} = time;
		return;
	}

	#No have previous status
	unless ($rdata->{servers}->{$_}->{status}){
		print "Previous Uptime is UNK\n" if ($config->{verbose});

		updateServerRRD($config,$server,$server_status);
		$rserver->{status} = $server_status;
		$rserver->{LastUpdate} = time;
		return;
	}


	if(didServerRestart($rserver->{status},$server_status)){
	  print "Server has been restarted\n" if ($config->{verbose});

		$rserver->{Restarts}++;

		updateServerRRD($config,$server,undef);
		$rserver->{status} = $server_status;
		$rserver->{LastUpdate} = time;
		return;
	}

	if (didFlashStatus($rdata->{servers}->{$_}->{status},$server_status)){
		print "Flush status detected\n" if ($config->{verbose});

    updateServerRRD($config,$server,undef);
    $rserver->{status} = $server_status;
		$rserver->{LastUpdate} = time;
    return;
	}

	updateServerRRD($config,$server,$server_status);

	if ($rserver->{status}){
	  $rdata->{Total_Questions} = 0 unless(defined($rdata->{Total_Questions}));
	  $rdata->{Total_In} = 0 unless(defined($rdata->{Total_In}));
	  $rdata->{Total_Out} = 0 unless(defined($rdata->{Total_Out}));

	  $rdata->{Total_Questions} += calculateDiff($rserver->{status}->{Questions},     $server_status->{Questions});
  	$rdata->{Total_In}				+= calculateDiff($rserver->{status}->{Bytes_received},$server_status->{Bytes_received});
	  $rdata->{Total_Out} 			+= calculateDiff($rserver->{status}->{Bytes_sent},    $server_status->{Bytes_sent});
	}

	$rserver->{status} = $server_status;
	$rserver->{LastUpdate} = time;

	print "\n" if ($config->{verbose});
}

###############################################################################
###############################################################################
sub printServerStatus{
	my $status = shift;

	if (defined($status)){
		print "Uptime: ".formatUptime($status->{Uptime})."\n";
		print "Questions: ".formatQuery($status->{Questions})."\n";
		print "Slow_queries: $status->{Slow_queries}\n";
		print "Connections: $status->{Connections}\n";
		print "Bytes_sent: ".formatTrafic($status->{Bytes_sent})."\n";
		print "Bytes_received: ".formatTrafic($status->{Bytes_received})."\n"
	}else{
		print "Uptime: UNK\n";
		print "Questions: UNK\n";
		print "Slow_queries: UNK\n";
		print "Connections: UNK\n";
		print "Bytes_sent: UNK\n";
		print "Bytes_received: UNK\n";
	}
}

###############################################################################
###############################################################################
sub didServerRestart{
	my $prev_status = shift;
	my $current_status = shift;

	return 1 if ($current_status->{Uptime}<$prev_status->{Uptime});
	return 1 if ($current_status->{Version} ne $prev_status->{Version});
}

###############################################################################
###############################################################################
sub didFlashStatus {
	my $prev_status = shift;
	my $current_status = shift;

	my ($c1,$c2) = (0,0);

	foreach(keys %{ $current_status }){
		next unless(/^(Handler|Com)_/);
		next if ($current_status->{$_}==$prev_status->{$_});

		$c1++;
	  $c2++ if ($current_status->{$_}<$prev_status->{$_});
	}

	return 0 unless($c1);

	return (($c2/$c1)*100)>80;
}

###############################################################################
###############################################################################
sub formatUptime {
	my $u = shift;

	my $sec = $u % 60;
	my $min = ($u / 60) % 60;
	my $hours = ($u / 3600) % 24;
	my $days = $u / (3600*24);

	my $str;

	if ($days){
		$str = sprintf("%d days %02d:%02d",$days,$hours,$min);
	}else{
		$str = sprintf("%02d:%02d",$hours,$min);
	}

	return $str;
}

###############################################################################
###############################################################################
sub formatTrafic {
	my $value = shift;

	if ($value > 1073741824){
		$value = sprintf('%0.2f Gb',$value/1073741824);
	}elsif ($value > 1048576){
		$value = sprintf('%0.2f Mb',$value/1048576);
	}elsif ($value > 1024){
		$value = sprintf('%0.2f Kb',$value/1024);
	}

	return $value
}

###############################################################################
###############################################################################
sub formatQuery {
	my $value = shift;

	if ($value > 1000000000){
		$value = sprintf('%0.2f GQ',$value/1000000000);
	}elsif ($value > 1000000){
		$value = sprintf('%0.2f MQ',$value/1000000);
	}elsif ($value > 1000){
		$value = sprintf('%0.2f KQ',$value/1000);
	}

	return $value
}

###############################################################################
###############################################################################
sub calculateDiff {
	my ($val1,$val2) = @_;

	return 0 if (!$val1 || !$val2 || $val1 eq 'UNK' || $val2 eq 'UNK');

	my $val = $val2-$val1;

	$val = 0xFFFFFFFF - $val1 + 1 + $val2 if($val<0);

	return $val;
}


sub printUsage {
	eval {
		&main::printUsage;
	}
}

sub printVersion {
	eval {
		&main::printVersion;
	}
}

1;
# vim:ts=2:sw=2
